--- old/usr/src/uts/common/fs/zfs/dmu_objset.c	2015-06-19 17:15:05.365046489 +0200
+++ new/usr/src/uts/common/fs/zfs/dmu_objset.c	2015-06-19 17:15:05.369036478 +0200
@@ -1746,7 +1746,19 @@
 	dmu_objset_find_ctx_t *dcp = arg;
 	dsl_pool_t *dp = dcp->dc_dp;
 
-	dsl_pool_config_enter(dp, FTAG);
+	/*
+	 * We need to get a pool_config_lock here, as there are several
+	 * asssert(pool_config_held) down the stack. Getting a lock via
+	 * dsl_pool_config_enter is risky, as it might be stalled by a
+	 * pending writer. This would deadlock, as the write lock can
+	 * only be granted when our parent thread gives up the lock.
+	 * The _prio interface gives us priority over a pending writer.
+	 * On the other hand, we don't risk to stall any pending writers,
+	 * as the parent thread already holds a config lock. We give up
+	 * our lock before the parent does, so in effect we do not prolong
+	 * the waiting time for the writer.
+	 */
+	dsl_pool_config_enter_prio(dp, FTAG);
 
 	dmu_objset_find_dp_impl(dcp);
 
--- old/usr/src/uts/common/fs/zfs/dsl_pool.c	2015-06-19 17:15:05.529898978 +0200
+++ new/usr/src/uts/common/fs/zfs/dsl_pool.c	2015-06-19 17:15:05.532412477 +0200
@@ -1046,6 +1046,13 @@
 }
 
 void
+dsl_pool_config_enter_prio(dsl_pool_t *dp, void *tag)
+{
+	ASSERT(!rrw_held(&dp->dp_config_rwlock, RW_READER));
+	rrw_enter_read_prio(&dp->dp_config_rwlock, tag);
+}
+
+void
 dsl_pool_config_exit(dsl_pool_t *dp, void *tag)
 {
 	rrw_exit(&dp->dp_config_rwlock, tag);
--- old/usr/src/uts/common/fs/zfs/rrwlock.c	2015-06-19 17:15:05.637485121 +0200
+++ new/usr/src/uts/common/fs/zfs/rrwlock.c	2015-06-19 17:15:05.639977214 +0200
@@ -159,8 +159,8 @@
 	refcount_destroy(&rrl->rr_linked_rcount);
 }
 
-void
-rrw_enter_read(rrwlock_t *rrl, void *tag)
+static void
+rrw_enter_read_impl(rrwlock_t *rrl, boolean_t prio, void *tag)
 {
 	mutex_enter(&rrl->rr_lock);
 #if !defined(DEBUG) && defined(_KERNEL)
@@ -176,7 +176,7 @@
 	ASSERT(refcount_count(&rrl->rr_anon_rcount) >= 0);
 
 	while (rrl->rr_writer != NULL || (rrl->rr_writer_wanted &&
-	    refcount_is_zero(&rrl->rr_anon_rcount) &&
+	    refcount_is_zero(&rrl->rr_anon_rcount) && !prio &&
 	    rrn_find(rrl) == NULL))
 		cv_wait(&rrl->rr_cv, &rrl->rr_lock);
 
@@ -192,6 +192,25 @@
 }
 
 void
+rrw_enter_read(rrwlock_t *rrl, void *tag)
+{
+	rrw_enter_read_impl(rrl, B_FALSE, tag);
+}
+
+/*
+ * take a read lock even if there are pending write lock requests. if we want
+ * to take a lock reentrantly, but from different threads (that have a
+ * relationship to each other), the normal detection mechanism to overrule
+ * the pending writer does not work, so we have to give an explicit hint here.
+ */
+void
+rrw_enter_read_prio(rrwlock_t *rrl, void *tag)
+{
+	rrw_enter_read_impl(rrl, B_TRUE, tag);
+}
+
+
+void
 rrw_enter_write(rrwlock_t *rrl)
 {
 	mutex_enter(&rrl->rr_lock);
--- old/usr/src/uts/common/fs/zfs/sys/dsl_pool.h	2015-06-19 17:15:05.737702389 +0200
+++ new/usr/src/uts/common/fs/zfs/sys/dsl_pool.h	2015-06-19 17:15:05.741218612 +0200
@@ -152,6 +152,7 @@
 void dsl_pool_mos_diduse_space(dsl_pool_t *dp,
     int64_t used, int64_t comp, int64_t uncomp);
 void dsl_pool_config_enter(dsl_pool_t *dp, void *tag);
+void dsl_pool_config_enter_prio(dsl_pool_t *dp, void *tag);
 void dsl_pool_config_exit(dsl_pool_t *dp, void *tag);
 boolean_t dsl_pool_config_held(dsl_pool_t *dp);
 boolean_t dsl_pool_config_held_writer(dsl_pool_t *dp);
--- old/usr/src/uts/common/fs/zfs/sys/rrwlock.h	2015-06-19 17:15:05.828801599 +0200
+++ new/usr/src/uts/common/fs/zfs/sys/rrwlock.h	2015-06-19 17:15:05.831278173 +0200
@@ -70,6 +70,7 @@
 void rrw_destroy(rrwlock_t *rrl);
 void rrw_enter(rrwlock_t *rrl, krw_t rw, void *tag);
 void rrw_enter_read(rrwlock_t *rrl, void *tag);
+void rrw_enter_read_prio(rrwlock_t *rrl, void *tag);
 void rrw_enter_write(rrwlock_t *rrl);
 void rrw_exit(rrwlock_t *rrl, void *tag);
 boolean_t rrw_held(rrwlock_t *rrl, krw_t rw);
